from .graph_spec import GraphSpec
from .roi import Roi
from .freezable import Freezable

import numpy as np
import networkx as nx

from copy import deepcopy
from typing import Dict, Optional, Set, Iterator, Any
import logging
import itertools


logger = logging.getLogger(__name__)


class Node(Freezable):
    """
    A stucture representing each node in a Graph.

    Args:

        id (``int``):

            A unique identifier for this Node

        location (``np.ndarray``):

            A numpy array containing a nodes location

        Optional attrs (``dict``, str -> ``Any``):

            A dictionary containing a mapping from attribute to value.
            Used to store any extra attributes associated with the
            Node such as color, size, etc.

        Optional temporary (bool):

            A tag to mark a node as temporary. Some operations such
            as `trim` might make new nodes that are just biproducts
            of viewing the data with a limited scope. These nodes
            are only guaranteed to have an id different from those
            in the same Graph, but may have conflicts if you request
            multiple graphs from the same source with different rois.
    """

    def __init__(
        self,
        id: int,
        location: np.ndarray,
        temporary: bool = False,
        attrs: Optional[Dict[str, Any]] = None,
    ):
        self.__attrs = attrs if attrs is not None else {}
        self.attrs["id"] = id
        self.location = location
        # purpose is to keep track of nodes that were created during
        # processing and do not have a corresponding node in the original source
        self.attrs["temporary"] = temporary
        self.freeze()

    def __getattr__(self, attr):
        if "__" not in attr:
            return self.attrs[attr]
        else:
            return super().__getattr__(attr)

    def __setattr__(self, attr, value):
        if "__" not in attr:
            self.attrs[attr] = value
        else:
            super().__setattr__(attr, value)

    @property
    def location(self):
        location = self.attrs["location"]
        return location

    @location.setter
    def location(self, new_location):
        assert isinstance(new_location, np.ndarray)
        self.attrs["location"] = new_location

    @property
    def id(self):
        return self.attrs["id"]

    @property
    def original_id(self):
        return self.id if not self.temporary else None

    @property
    def temporary(self):
        return self.attrs["temporary"]

    @property
    def attrs(self):
        return self.__attrs

    @property
    def all(self):
        return self.attrs

    @classmethod
    def from_attrs(cls, attrs: Dict[str, Any]):
        node_id = attrs["id"]
        location = attrs["location"]
        temporary = attrs.get("temporary", False)
        return cls(id=node_id, location=location, temporary=temporary, attrs=attrs)

    def __str__(self):
        return f"Node({self.temporary}) ({self.id}) at ({self.location})"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return isinstance(other, Node) and self.id == other.id

    def __hash__(self):
        return hash(self.id)


class Edge(Freezable):
    """
    A structure representing edges in a graph.

    Args:

        u (``int``)

            The id of the 'u' node of this edge

        v (``int``)

            the id of the `v` node of this edge
    """

    def __init__(self, u: int, v: int, attrs: Optional[Dict[str, Any]] = None):
        self.__u = u
        self.__v = v
        self.__attrs = attrs if attrs is not None else {}
        self.freeze()

    @property
    def u(self):
        return self.__u

    @property
    def v(self):
        return self.__v

    @property
    def all(self):
        return self.__attrs

    def __iter__(self):
        return iter([self.u, self.v])

    def __str__(self):
        return f"({self.u}, {self.v})"

    def __repr__(self):
        return f"({self.u}, {self.v})"

    def __eq__(self, other):
        return self.u == other.u and self.v == other.v

    def __hash__(self):
        return hash((self.u, self.v))

    def directed_eq(self, other):
        return self.u == other.u and self.v == other.v

    def undirected_eq(self, other):
        return set([self.u, self.v]) == set([other.u, other.v])


class Graph(Freezable):
    """A structure containing a list of :class:`Node`, a list of :class:`Edge`,
    and a specification describing the data.

    Args:

        nodes (``iterator``, :class:`Node`):

            An iterator containing Vertices.

        edges (``iterator``, :class:`Edge`):

            An iterator containing Edges.

        spec (:class:`GraphSpec`):

            A spec describing the data.
    """

    def __init__(self, nodes: Iterator[Node], edges: Iterator[Edge], spec: GraphSpec):
        self.__spec = spec
        self.__graph = self.create_graph(nodes, edges)

    @property
    def spec(self):
        return self.__spec

    @spec.setter
    def spec(self, new_spec):
        self.__spec = new_spec

    @property
    def directed(self):
        return (
            self.spec.directed
            if self.spec.directed is not None
            else self.__graph.is_directed()
        )

    def create_graph(self, nodes: Iterator[Node], edges: Iterator[Edge]):
        if self.__spec.directed is None:
            logger.debug(
                "Trying to create a Graph without specifying directionality. Using default Directed!"
            )
            graph = nx.DiGraph()
        elif self.__spec.directed:
            graph = nx.DiGraph()
        else:
            graph = nx.Graph()

        for node in nodes:
            node.location = node.location.astype(self.spec.dtype)

        vs = [(v.id, v.all) for v in nodes]
        graph.add_nodes_from(vs)
        graph.add_edges_from([(e.u, e.v, e.all) for e in edges])
        return graph

    @property
    def nodes(self):
        for node_id, node_attrs in self.__graph.nodes.items():
            if "id" not in node_attrs:
                node_attrs["id"] = node_id
            v = Node.from_attrs(node_attrs)
            if not np.issubdtype(v.location.dtype, self.spec.dtype):
                raise Exception(
                    f"expected location to have dtype {self.spec.dtype} but it had {v.location.dtype}"
                )
            yield v

    def num_vertices(self):
        return self.__graph.number_of_nodes()

    def num_edges(self):
        return self.__graph.number_of_edges()

    @property
    def edges(self):
        for (u, v), attrs in self.__graph.edges.items():
            yield Edge(u, v, attrs)

    def neighbors(self, node):
        if self.directed:
            for neighbor in self.__graph.successors(node.id):
                yield Node.from_attrs(self.__graph.nodes[neighbor])
            if self.directed:
                for neighbor in self.__graph.predecessors(node.id):
                    yield Node.from_attrs(self.__graph.nodes[neighbor])
        else:
            for neighbor in self.__graph.neighbors(node.id):
                yield Node.from_attrs(self.__graph.nodes[neighbor])

    def __str__(self):
        string = "Vertices:\n"
        for node in self.nodes:
            string += f"{node}\n"
        string += "Edges:\n"
        for edge in self.edges:
            string += f"{edge}\n"
        return string

    def __repr__(self):
        return str(self)

    def node(self, id: int):
        """
        Get node with a specific id
        """
        attrs = self.__graph.nodes[id]
        return Node.from_attrs(attrs)

    def contains(self, node_id: int):
        return node_id in self.__graph.nodes

    def remove_node(self, node: Node, retain_connectivity=False):
        """
        Remove a node.

        retain_connectivity: preserve removed nodes neighboring edges.
        Given graph: a->b->c, removing `b` without retain_connectivity
        would leave us with two connected components, {'a'} and {'b'}.
        removing 'b' with retain_connectivity flag set to True would
        leave us with the graph: a->c, and only one connected component
        {a, c}, thus preserving the connectivity of 'a' and 'c'
        """
        if retain_connectivity:
            predecessors = self.predecessors(node)
            successors = self.successors(node)

            for pred_id in predecessors:
                for succ_id in successors:
                    if pred_id != succ_id:
                        self.add_edge(Edge(pred_id, succ_id))
        self.__graph.remove_node(node.id)

    def add_node(self, node: Node):
        """
        Adds a node to the graph.
        If a node exists with the same id as the node you are adding,
        its attributes will be overwritten.
        """
        node.location = node.location.astype(self.spec.dtype)
        self.__graph.add_node(node.id, **node.all)

    def remove_edge(self, edge: Edge):
        """
        Remove an edge from the graph.
        """
        self.__graph.remove_edge(edge.u, edge.v)

    def add_edge(self, edge: Edge):
        """
        Adds an edge to the graph.
        If an edge exists with the same u and v, its attributes
        will be overwritten.
        """
        self.__graph.add_edge(edge.u, edge.v, **edge.all)

    def copy(self):
        return deepcopy(self)

    def crop(self, roi: Roi):
        """
        Will remove all nodes from self that are not contained in `roi` except for
        "dangling" nodes. This means that if there are nodes A, B s.t. there
        is an edge (A, B) and A is contained in `roi` but B is not, the edge (A, B)
        is considered contained in the `roi` and thus node B will be kept as a
        "dangling" node.

        Note there is a helper function `trim` that will remove B and replace it with
        a node at the intersection of the edge (A, B) and the bounding box of `roi`.

        Args:

            roi (:class:`Roi`):

                ROI in world units to crop to.
        """

        cropped = self.copy()

        contained_nodes = set([v.id for v in cropped.nodes if roi.contains(v.location)])
        all_contained_edges = set(
            [
                e
                for e in cropped.edges
                if e.u in contained_nodes or e.v in contained_nodes
            ]
        )
        fully_contained_edges = set(
            [
                e
                for e in all_contained_edges
                if e.u in contained_nodes and e.v in contained_nodes
            ]
        )
        partially_contained_edges = all_contained_edges - fully_contained_edges
        contained_edge_nodes = set(list(itertools.chain(*all_contained_edges)))
        all_nodes = contained_edge_nodes | contained_nodes
        dangling_nodes = all_nodes - contained_nodes

        for node in list(cropped.nodes):
            if node.id not in all_nodes:
                cropped.remove_node(node)
        for edge in list(cropped.edges):
            if edge not in all_contained_edges:
                cropped.remove_edge(edge)

        cropped.spec.roi = roi
        return cropped

    def shift(self, offset):
        for node in self.nodes:
            node.location += offset

    def new_graph(self):
        if self.directed():
            return nx.DiGraph()
        else:
            return nx.Graph()

    def trim(self, roi: Roi):
        """
        Create a copy of self and replace "dangling" nodes with contained nodes.

        A "dangling" node is defined by: Let A, B be nodes s.t. there exists an
        edge (A, B) and A is contained in `roi` but B is not. Edge (A, B) is considered
        contained, and thus B is kept as a "dangling" node.
        """

        trimmed = self.copy()

        contained_nodes = set([v.id for v in trimmed.nodes if roi.contains(v.location)])
        all_contained_edges = set(
            [
                e
                for e in trimmed.edges
                if e.u in contained_nodes or e.v in contained_nodes
            ]
        )
        fully_contained_edges = set(
            [
                e
                for e in all_contained_edges
                if e.u in contained_nodes and e.v in contained_nodes
            ]
        )
        partially_contained_edges = all_contained_edges - fully_contained_edges
        contained_edge_nodes = set(list(itertools.chain(*all_contained_edges)))
        all_nodes = contained_edge_nodes | contained_nodes
        dangling_nodes = all_nodes - contained_nodes

        next_node = 0 if len(all_nodes) == 0 else max(all_nodes) + 1

        trimmed._handle_boundaries(
            partially_contained_edges,
            contained_nodes,
            roi,
            node_id=itertools.count(next_node),
        )

        for node in trimmed.nodes:
            assert roi.contains(
                node.location
            ), f"Failed to properly contain node {node.id} at {node.location}"

        return trimmed

    def _handle_boundaries(
        self,
        crossing_edges: Iterator[Edge],
        contained_nodes: Set[int],
        roi: Roi,
        node_id: Iterator[int],
    ):
        nodes_to_remove = set([])
        for e in crossing_edges:
            u, v = self.node(e.u), self.node(e.v)
            u_in = u.id in contained_nodes
            v_in, v_out = (u, v) if u_in else (v, u)
            in_location, out_location = (v_in.location, v_out.location)
            new_location = self._roi_intercept(in_location, out_location, roi)
            if not all(np.isclose(new_location, in_location)):
                # use deepcopy because modifying this node should not modify original
                new_attrs = deepcopy(v_out.attrs)
                new_attrs["id"] = next(node_id)
                new_attrs["location"] = new_location
                new_attrs["temporary"] = True
                new_v = Node.from_attrs(new_attrs)
                new_e = Edge(
                    u=v_in.id if u_in else new_v.id, v=new_v.id if u_in else v_in.id
                )
                self.add_node(new_v)
                self.add_edge(new_e)
            nodes_to_remove.add(v_out)
        for node in nodes_to_remove:
            self.remove_node(node)

    def _roi_intercept(
        self, inside: np.ndarray, outside: np.ndarray, bb: Roi
    ) -> np.ndarray:
        """
        Given two points, one inside a bounding box and one outside,
        get the intercept between the line and the bounding box.
        """

        offset = outside - inside
        distance = np.linalg.norm(offset)
        assert not np.isclose(distance, 0), "Inside and Outside are the same location"
        direction = offset / distance

        # `offset` can be 0 on some but not all axes leaving a 0 in the denominator.
        # `inside` can be on the bounding box, leaving a 0 in the numerator.
        # `x/0` throws a division warning, `0/0` throws an invalid warning (both are fine here)
        with np.errstate(divide="ignore", invalid="ignore"):
            bb_x = np.asarray(
                [
                    (np.asarray(bb.begin) - inside) / offset,
                    (np.asarray(bb.end) - inside) / offset,
                ],
                dtype=self.spec.dtype,
            )

        with np.errstate(invalid="ignore"):
            s = np.min(bb_x[np.logical_and((bb_x >= 0), (bb_x <= 1))])

        new_location = inside + s * distance * direction
        upper = np.array(bb.end, dtype=self.spec.dtype)
        new_location = np.clip(
            new_location, bb.begin, upper - upper * np.finfo(self.spec.dtype).eps
        )
        return new_location

    def merge(self, other, copy_from_self=False, copy=False):
        """
        Merge this graph with another. The resulting graph will have the Roi
        of the larger one.

        This only works if one of the two graphs contains the other.
        In this case, ``other`` will overwrite edges and nodes with the same
        ID in ``self`` (unless ``copy_from_self`` is set to ``True``).
        Vertices and edges in ``self`` that are contained in the Roi of ``other``
        will be removed (vice versa for ``copy_from_self``)

        A copy will only be made if necessary or ``copy`` is set to ``True``.
        """

        # It is unclear how to merge points in all cases. Consider a 10x10 graph,
        # you crop out a 5x5 area, do a shift augment, and attempt to merge.
        # What does that mean? specs have changed. It should be a new key.
        raise NotImplementedError("Merge function should not be used!")

        self_roi = self.spec.roi
        other_roi = other.spec.roi

        assert self_roi.contains(other_roi) or other_roi.contains(
            self_roi
        ), "Can not merge graphs that are not contained in each other."

        # make sure self contains other
        if not self_roi.contains(other_roi):
            return other.merge(self, not copy_from_self, copy)

        # edges and nodes in addition are guaranteed to be in merged
        base = other if copy_from_self else self
        addition = self if copy_from_self else other

        if copy:
            merged = base.copy()
        else:
            merged = base

        for node in list(merged.nodes):
            if merged.spec.roi.contains(node.location):
                merged.remove_node(node)
        for edge in list(merged.edges):
            if merged.spec.roi.contains(
                merged.node(edge.u)
            ) or merged.spec.roi.contains(merged.node(edge.v)):
                merged.remove_edge(edge)
        for node in addition.nodes:
            merged.add_node(node)
        for edge in addition.edges:
            merged.add_edge(edge)

        return merged

    def to_nx_graph(self):
        """
        returns a pure networkx graph containing data from
        this Graph.
        """
        return deepcopy(self.__graph)

    @classmethod
    def from_nx_graph(cls, graph, spec):
        """
        Create a gunpowder graph from a networkx graph.
        The network graph is expected to have a "location"
        attribute for each node. If it is a subclass of a networkx
        graph with extra functionality, this may not work.
        """
        if spec.directed is None:
            spec.directed = graph.is_directed()
        g = cls([], [], spec)
        g.__graph = graph
        return g

    def relabel_connected_components(self):
        """
        create a new attribute "component" for each node
        in this Graph
        """
        for i, wcc in enumerate(self.connected_components):
            for node in wcc:
                self.__graph.nodes[node]["component"] = i

    @property
    def connected_components(self):
        if not self.directed:
            return nx.connected_components(self.__graph)
        else:
            return nx.weakly_connected_components(self.__graph)

    def in_degree(self):
        return self.__graph.in_degree()

    def successors(self, node):
        if self.directed:
            return self.__graph.successors(node.id)
        else:
            return self.__graph.neighbors(node.id)

    def predecessors(self, node):
        if self.directed:
            return self.__graph.predecessors(node.id)
        else:
            return self.__graph.neighbors(node.id)


class GraphKey(Freezable):
    """A key to identify graphs in requests, batches, and across
    nodes.

    Used as key in :class:`BatchRequest` and :class:`Batch` to retrieve specs
    or graphs.

    Args:

        identifier (``string``):

            A unique, human readable identifier for this graph key. Will be
            used in log messages and to look up graphs in requests and batches.
            Should be upper case (like ``CENTER_GRAPH``). The identifier is
            unique: Two graph keys with the same identifier will refer to the
            same graph.
    """

    def __init__(self, identifier):
        self.identifier = identifier
        self.hash = hash(identifier)
        self.freeze()
        logger.debug("Registering graph type %s", self)
        setattr(GraphKeys, self.identifier, self)

    def __eq__(self, other):
        return hasattr(other, "identifier") and self.identifier == other.identifier

    def __hash__(self):
        return self.hash

    def __repr__(self):
        return self.identifier


class GraphKeys:
    """Convenience access to all created :class:`GraphKey`s. A key generated
    with::

        centers = GraphKey('CENTER_GRAPH')

    can be retrieved as::

        GraphKeys.CENTER_GRAPH
    """

    pass



from __future__ import absolute_import

from .nodes import *

from .array import Array, ArrayKey, ArrayKeys
from .array_spec import ArraySpec
from .batch import Batch
from .batch_request import BatchRequest
from .build import build
from .coordinate import Coordinate
from .graph import Graph, Node, Edge, GraphKey, GraphKeys
from .graph_spec import GraphSpec
from .pipeline import *
from .producer_pool import ProducerPool
from .provider_spec import ProviderSpec
from .roi import Roi
from .version_info import _version as version
import gunpowder.contrib
import gunpowder.tensorflow
import gunpowder.torch
import gunpowder.jax
import gunpowder.zoo



from random import randrange
import logging
import numpy as np

from gunpowder.coordinate import Coordinate

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class SpecifiedLocation(BatchFilter):
    """Choses a batch at a location from the list provided at init, making sure
    it is in the bounding box of the upstream provider.

    Locations should be given in world units.

    Locations will be chosen in order or at random from the list depending on the
    ``choose_randomly`` parameter.

    If a location requires a shift outside the bounding box of any upstream provider
    the module will skip that location with a warning.

    Args:

        locations (``list`` of locations):

            Locations to center batches around.

        choose_randomly (``bool``):

            Defines whether locations should be picked in order or at random
            from the list.

        extra_data (``list`` of array-like):

            A list of data that will be passed along with the arrays provided
            by this node. This data will be appended as an attribute to the
            dataset so it must be a data format compatible with hdf5.

        jitter (``tuple`` of int):

            How far to allow the point to shift in each direction.
            Default is None, which places the point in the center.
            Chooses uniformly from [loc - jitter, loc + jitter] in each
            direction.
    """

    def __init__(self, locations, choose_randomly=False, extra_data=None, jitter=None):
        self.coordinates = locations
        self.choose_randomly = choose_randomly
        self.jitter = jitter
        self.loc_i = -1
        self.upstream_spec = None
        self.specified_shift = None

        if extra_data is not None:
            assert len(extra_data) == len(locations), (
                "extra_data (%d) should match the length of specified locations (%d)"
                % (len(extra_data), len(locations))
            )

        self.extra_data = extra_data

    def setup(self):
        self.upstream_spec = self.get_upstream_provider().spec

        # clear bounding boxes of all provided arrays and points --
        # SpecifiedLocation does know its locations at setup (checks on the fly)
        for key, spec in self.spec.items():
            spec.roi.shape = (None,) * spec.roi.dims
            self.updates(key, spec)

    def prepare(self, request):
        lcm_voxel_size = self.spec.get_lcm_voxel_size(request.array_specs.keys())

        # shift to center
        total_roi = request.get_total_roi()
        request_center = total_roi.shape / 2 + total_roi.offset

        self.specified_shift = self._get_next_shift(request_center, lcm_voxel_size)
        while not self.__check_shift(request):
            logger.warning(
                "Location %s (shift %s) skipped"
                % (self.coordinates[self.loc_i], self.specified_shift)
            )
            self.specified_shift = self._get_next_shift(request_center, lcm_voxel_size)

        # Set shift for all requests
        for specs_type in [request.array_specs, request.graph_specs]:
            for key, spec in specs_type.items():
                roi = spec.roi.shift(self.specified_shift)
                specs_type[key].roi = roi

        logger.debug(
            "{}'th ({}) shift selected: {}".format(
                self.loc_i, self.coordinates[self.loc_i], self.specified_shift
            )
        )

        deps = request
        return deps

    def process(self, batch, request):
        # reset ROIs to request
        for array_key, spec in request.array_specs.items():
            batch.arrays[array_key].spec.roi = spec.roi
            if self.extra_data is not None:
                batch.arrays[array_key].attrs[
                    "specified_location_extra_data"
                ] = self.extra_data[self.loc_i]

        for graph_key, spec in request.graph_specs.items():
            batch.points[graph_key].spec.roi = spec.roi

        # change shift point locations to lie within roi
        for graph_key in request.graph_specs.keys():
            batch.points[graph_key].shift(-self.specified_shift)

    def _get_next_shift(self, center_shift, voxel_size):
        # gets next coordinate from list

        if self.choose_randomly:
            self.loc_i = randrange(len(self.coordinates))
        else:
            self.loc_i += 1
            if self.loc_i >= len(self.coordinates):
                self.loc_i = 0
                logger.warning("Ran out of specified locations, looping list")
        next_shift = Coordinate(self.coordinates[self.loc_i]) - center_shift

        if self.jitter is not None:
            rnd = []
            for i in range(len(self.jitter)):
                rnd.append(np.random.randint(-self.jitter[i], self.jitter[i] + 1))
            next_shift += Coordinate(rnd)
        logger.debug("Shift before rounding: %s" % str(next_shift))
        # make sure shift is a multiple of voxel size (round to nearest)
        next_shift = Coordinate(
            [
                int(vs * round(float(shift) / vs))
                for vs, shift in zip(voxel_size, next_shift)
            ]
        )
        logger.debug("Shift after rounding: %s" % str(next_shift))
        return next_shift

    def __check_shift(self, request):
        for key, spec in request.items():
            request_roi = spec.roi
            if key in self.upstream_spec:
                provided_roi = self.upstream_spec[key].roi
            else:
                raise Exception("Requested %s, but upstream does not provide it." % key)
            shifted_roi = request_roi.shift(self.specified_shift)
            if not provided_roi.contains(shifted_roi):
                logger.warning(
                    "Provided roi %s for key %s does not contain shifted roi %s"
                    % (provided_roi, key, shifted_roi)
                )
                return False
        return True



from .batch_filter import BatchFilter
from gunpowder.array import ArrayKey, Array
from gunpowder.batch_request import BatchRequest
from gunpowder.batch import Batch
import logging
import numbers
from funlib.geometry import Coordinate

logger = logging.getLogger(__name__)


class DownSample(BatchFilter):
    """Downsample arrays in a batch by given factors.

    Args:

        source (:class:`ArrayKey`):

            The key of the array to downsample.

        factor (``int`` or ``tuple`` of ``int``):

            The factor to downsample with.

        target (:class:`ArrayKey`):

            The key of the array to store the downsampled ``source``.
    """

    def __init__(self, source, factor, target):
        assert isinstance(source, ArrayKey)
        assert isinstance(target, ArrayKey)
        assert isinstance(factor, numbers.Number) or isinstance(
            factor, tuple
        ), "Scaling factor should be a number or a tuple of numbers."

        self.source = source
        self.factor = (
            factor if isinstance(factor, numbers.Number) else Coordinate(factor)
        )
        self.target = target

    def setup(self):
        spec = self.spec[self.source].copy()
        spec.voxel_size *= self.factor
        self.provides(self.target, spec)
        self.enable_autoskip()

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.source] = request[self.target]
        return deps

    def process(self, batch, request):
        outputs = Batch()
        data = batch.arrays[self.source].data

        channel_dims = len(data.shape) - batch.arrays[self.source].spec.roi.dims

        # downsample
        if isinstance(self.factor, tuple):
            slices = tuple(slice(None, None) for _ in range(channel_dims)) + tuple(
                slice(None, None, k) for k in self.factor
            )
        else:
            slices = tuple(slice(None, None) for _ in range(channel_dims)) + tuple(
                slice(None, None, self.factor)
                for i in range(batch[self.source].spec.roi.dims)
            )

        logger.debug("downsampling %s with %s", self.source, slices)

        data = data[slices]

        # create output array
        spec = self.spec[self.target].copy()
        spec.roi = request[self.target].roi
        outputs.arrays[self.target] = Array(data, spec)

        return outputs



from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.profiling import Timing
from gunpowder.roi import Roi
from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from .batch_provider import BatchProvider

import logging
import numpy as np
import warnings

logger = logging.getLogger(__name__)


class Hdf5LikeSource(BatchProvider):
    """An HDF5-like data source.

    Provides arrays from datasets accessed with an h5py-like API for each array
    key given. If the attribute ``resolution`` is set in a dataset, it will be
    used as the array's ``voxel_size``. If the attribute ``offset`` is set in a
    dataset, it will be used as the offset of the :class:`Roi` for this array.
    It is assumed that the offset is given in world units.

    Args:

        filename (``string``):

            The input file.

        datasets (``dict``, :class:`ArrayKey` -> ``string``):

            Dictionary of array keys to dataset names that this source offers.

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            An optional dictionary of array keys to array specs to overwrite
            the array specs automatically determined from the data file. This
            is useful to set a missing ``voxel_size``, for example. Only fields
            that are not ``None`` in the given :class:`ArraySpec` will be used.

        channels_first (``bool``, optional):

            Specifies the ordering of the dimensions of the HDF5-like data source.
            If channels_first is set (default), then the input shape is expected
            to be (channels, spatial dimensions). This is recommended due to
            better performance. If channels_first is set to false, then the input
            data is read in channels_last manner and converted to channels_first.
    """

    def __init__(self, filename, datasets, array_specs=None, channels_first=True):
        warnings.warn(
            "HDF5LikeSource is depricated and will soon be removed in v2.0",
            DeprecationWarning,
        )

        self.filename = filename
        self.datasets = datasets

        if array_specs is None:
            self.array_specs = {}
        else:
            self.array_specs = array_specs

        self.channels_first = channels_first

        # number of spatial dimensions
        self.ndims = None

    def _open_file(self, filename):
        raise NotImplementedError("Only implemented in subclasses")

    def setup(self):
        with self._open_file(self.filename) as data_file:
            for array_key, ds_name in self.datasets.items():
                if ds_name not in data_file:
                    raise RuntimeError("%s not in %s" % (ds_name, self.filename))

                spec = self.__read_spec(array_key, data_file, ds_name)

                self.provides(array_key, spec)

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        batch = Batch()

        with self._open_file(self.filename) as data_file:
            for array_key, request_spec in request.array_specs.items():
                logger.debug("Reading %s in %s...", array_key, request_spec.roi)

                voxel_size = self.spec[array_key].voxel_size

                # scale request roi to voxel units
                dataset_roi = request_spec.roi / voxel_size

                # shift request roi into dataset
                dataset_roi = dataset_roi - self.spec[array_key].roi.offset / voxel_size

                # create array spec
                array_spec = self.spec[array_key].copy()
                array_spec.roi = request_spec.roi

                # add array to batch
                batch.arrays[array_key] = Array(
                    self.__read(data_file, self.datasets[array_key], dataset_roi),
                    array_spec,
                )

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def _get_voxel_size(self, dataset):
        try:
            return Coordinate(dataset.attrs["resolution"])
        except Exception:  # todo: make specific when z5py supports it
            return None

    def _get_offset(self, dataset):
        try:
            return Coordinate(dataset.attrs["offset"])
        except Exception:  # todo: make specific when z5py supports it
            return None

    def __read_spec(self, array_key, data_file, ds_name):
        dataset = data_file[ds_name]

        if array_key in self.array_specs:
            spec = self.array_specs[array_key].copy()
        else:
            spec = ArraySpec()

        if spec.voxel_size is None:
            voxel_size = self._get_voxel_size(dataset)
            if voxel_size is None:
                voxel_size = Coordinate((1,) * len(dataset.shape))
                logger.warning(
                    "WARNING: File %s does not contain resolution information "
                    "for %s (dataset %s), voxel size has been set to %s. This "
                    "might not be what you want.",
                    self.filename,
                    array_key,
                    ds_name,
                    spec.voxel_size,
                )
            spec.voxel_size = voxel_size

        self.ndims = len(spec.voxel_size)

        if spec.roi is None:
            offset = self._get_offset(dataset)
            if offset is None:
                offset = Coordinate((0,) * self.ndims)

            if self.channels_first:
                shape = Coordinate(dataset.shape[-self.ndims :])
            else:
                shape = Coordinate(dataset.shape[: self.ndims])

            spec.roi = Roi(offset, shape * spec.voxel_size)

        if spec.dtype is not None:
            assert spec.dtype == dataset.dtype, (
                "dtype %s provided in array_specs for %s, "
                "but differs from dataset %s dtype %s"
                % (self.array_specs[array_key].dtype, array_key, ds_name, dataset.dtype)
            )
        else:
            spec.dtype = dataset.dtype

        if spec.interpolatable is None:
            spec.interpolatable = spec.dtype in [
                np.float32,
                np.float64,
                np.float128,
                np.uint8,  # assuming this is not used for labels
            ]
            logger.warning(
                "WARNING: You didn't set 'interpolatable' for %s "
                "(dataset %s). Based on the dtype %s, it has been "
                "set to %s. This might not be what you want.",
                array_key,
                ds_name,
                spec.dtype,
                spec.interpolatable,
            )

        return spec

    def __read(self, data_file, ds_name, roi):
        c = len(data_file[ds_name].shape) - self.ndims

        if self.channels_first:
            array = np.asarray(data_file[ds_name][(slice(None),) * c + roi.to_slices()])
        else:
            array = np.asarray(data_file[ds_name][roi.to_slices() + (slice(None),) * c])
            array = np.transpose(
                array, axes=[i + self.ndims for i in range(c)] + list(range(self.ndims))
            )

        return array

    def name(self):
        return super().name() + f"[{self.filename}]"



import logging
import numpy as np

from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.ext import dvision
from gunpowder.profiling import Timing
from gunpowder.roi import Roi
from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from .batch_provider import BatchProvider

logger = logging.getLogger(__name__)


class DvidSource(BatchProvider):
    """A DVID array source.

    Provides arrays from DVID servers for each array key given.

    Args:

        hostname (``string``):

            The name of the DVID server.

        port (``int``):

            The port of the DVID server.

        uuid (``string``):

            The UUID of the DVID node to use.

        datasets (``dict``, :class:`ArrayKey` -> ``string``):

            Dictionary mapping array keys to DVID data instance names that this
            source offers.

        masks (``dict``, :class:`ArrayKey` -> ``string``, optional):

            Dictionary of array keys to DVID ROI instance names. This will
            create binary masks from DVID ROIs.

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            An optional dictionary of array keys to specs to overwrite the
            array specs automatically determined from the DVID server. This is
            useful to set ``voxel_size``, for example. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.
    """

    def __init__(self, hostname, port, uuid, datasets, masks=None, array_specs=None):
        self.hostname = hostname
        self.port = port
        self.url = "http://{}:{}".format(self.hostname, self.port)
        self.uuid = uuid

        self.datasets = datasets
        self.masks = masks if masks is not None else {}

        self.array_specs = array_specs if array_specs is not None else {}

        self.ndims = None

    def setup(self):
        for array_key, _ in self.datasets.items():
            spec = self.__get_spec(array_key)
            self.provides(array_key, spec)

        for array_key, _ in self.masks.items():
            spec = self.__get_mask_spec(array_key)
            self.provides(array_key, spec)

        logger.info("DvidSource.spec:\n%s", self.spec)

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        batch = Batch()

        for array_key, request_spec in request.array_specs.items():
            logger.debug("Reading %s in %s...", array_key, request_spec.roi)

            voxel_size = self.spec[array_key].voxel_size

            # scale request roi to voxel units
            dataset_roi = request_spec.roi / voxel_size

            # shift request roi into dataset
            dataset_roi = dataset_roi - self.spec[array_key].roi.offset / voxel_size

            # create array spec
            array_spec = self.spec[array_key].copy()
            array_spec.roi = request_spec.roi

            # read the data
            if array_key in self.datasets:
                data = self.__read_array(self.datasets[array_key], dataset_roi)
            elif array_key in self.masks:
                data = self.__read_mask(self.masks[array_key], dataset_roi)
            else:
                assert False, (
                    "Encountered a request for %s that is neither a volume "
                    "nor a mask." % array_key
                )

            # add array to batch
            batch.arrays[array_key] = Array(data, array_spec)

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __get_info(self, array_key):
        if array_key in self.datasets:
            data = dvision.DVIDDataInstance(
                self.hostname, self.port, self.uuid, self.datasets[array_key]
            )

        elif array_key in self.masks:
            data = dvision.DVIDRegionOfInterest(
                self.hostname, self.port, self.uuid, self.masks[array_key]
            )

        else:
            assert False, (
                "Encountered a request that is neither a volume " "nor a mask."
            )

        return data.info

    def __get_spec(self, array_key):
        info = self.__get_info(array_key)

        roi_min = info["Extended"]["MinPoint"]
        if roi_min is not None:
            roi_min = Coordinate(roi_min[::-1])
        roi_max = info["Extended"]["MaxPoint"]
        if roi_max is not None:
            roi_max = Coordinate(roi_max[::-1])

        data_roi = Roi(offset=roi_min, shape=(roi_max - roi_min))
        data_dims = Coordinate(data_roi.shape)

        if self.ndims is None:
            self.ndims = len(data_dims)
        else:
            assert self.ndims == len(data_dims)

        if array_key in self.array_specs:
            spec = self.array_specs[array_key].copy()
        else:
            spec = ArraySpec()

        if spec.voxel_size is None:
            spec.voxel_size = Coordinate(info["Extended"]["VoxelSize"])

        if spec.roi is None:
            spec.roi = data_roi * spec.voxel_size

        data_dtype = dvision.DVIDDataInstance(
            self.hostname, self.port, self.uuid, self.datasets[array_key]
        ).dtype

        if spec.dtype is not None:
            assert spec.dtype == data_dtype, (
                "dtype %s provided in array_specs for %s, "
                "but differs from instance %s dtype %s"
                % (
                    self.array_specs[array_key].dtype,
                    array_key,
                    self.datasets[array_key],
                    data_dtype,
                )
            )
        else:
            spec.dtype = data_dtype

        if spec.interpolatable is None:
            spec.interpolatable = spec.dtype in [
                np.float32,
                np.float64,
                np.float128,
                np.uint8,  # assuming this is not used for labels
            ]
            logger.warning(
                "WARNING: You didn't set 'interpolatable' for %s. "
                "Based on the dtype %s, it has been set to %s. "
                "This might not be what you want.",
                array_key,
                spec.dtype,
                spec.interpolatable,
            )

        return spec

    def __get_mask_spec(self, mask_key):
        # create initial array spec

        if mask_key in self.array_specs:
            spec = self.array_specs[mask_key].copy()
        else:
            spec = ArraySpec()

        # get voxel size

        if spec.voxel_size is None:
            voxel_size = None
            for array_key in self.datasets:
                if voxel_size is None:
                    voxel_size = self.spec[array_key].voxel_size
                else:
                    assert voxel_size == self.spec[array_key].voxel_size, (
                        "No voxel size was given for mask %s, and the voxel "
                        "sizes of the volumes %s are not all the same. I don't "
                        "know what voxel size to use to create the mask."
                        % (mask_key, self.datasets.keys())
                    )

            spec.voxel_size = voxel_size

        # get ROI

        if spec.roi is None:
            for array_key in self.datasets:
                roi = self.spec[array_key].roi

                if spec.roi is None:
                    spec.roi = roi.copy()
                else:
                    spec.roi = roi.union(spec.roi)

        # set interpolatable

        if spec.interpolatable is None:
            spec.interpolatable = False

        # set datatype

        if spec.dtype is not None and spec.dtype != np.uint8:
            logger.warn(
                "Ignoring dtype in array_spec for %s, only np.uint8 "
                "is allowed for masks.",
                mask_key,
            )
        spec.dtype = np.uint8

        return spec

    def __read_array(self, instance, roi):
        data_instance = dvision.DVIDDataInstance(
            self.hostname, self.port, self.uuid, instance
        )

        return data_instance[roi.get_bounding_box()]

    def __read_mask(self, instance, roi):
        dvid_roi = dvision.DVIDRegionOfInterest(
            self.hostname, self.port, self.uuid, instance
        )

        return dvid_roi[roi.get_bounding_box()]

    def __repr__(self):
        return "DvidSource(hostname={}, port={}, uuid={}".format(
            self.hostname, self.port, self.uuid
        )



import logging
import random
import numpy as np

# imports for deformed slice
from skimage.draw import line
from scipy.ndimage.measurements import label
from scipy.ndimage.interpolation import map_coordinates
from scipy.ndimage.morphology import binary_dilation

from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class DefectAugment(BatchFilter):
    """Augment intensity arrays section-wise with artifacts like missing
    sections, low-contrast sections, by blending in artifacts drawn from a
    separate source, or by deforming a section.

    Args:

        intensities (:class:`ArrayKey`):

            The key of the array of intensities to modify.

        prob_missing(``float``):
        prob_low_contrast(``float``):
        prob_artifact(``float``):
        prob_deform(``float``):

            Probabilities of having a missing section, low-contrast section, an
            artifact (see param ``artifact_source``) or a deformed slice. The
            sum should not exceed 1. Values in missing sections will be set to
            0.

        contrast_scale (``float``, optional):

            By how much to scale the intensities for a low-contrast section,
            used if ``prob_low_contrast`` > 0.

        artifact_source (class:`BatchProvider`, optional):

            A gunpowder batch provider that delivers intensities (via
            :class:`ArrayKey` ``artifacts``) and an alpha mask (via
            :class:`ArrayKey` ``artifacts_mask``), used if ``prob_artifact`` > 0.

        artifacts(:class:`ArrayKey`, optional):

            The key to query ``artifact_source`` for to get the intensities
            of the artifacts.

        artifacts_mask(:class:`ArrayKey`, optional):

            The key to query ``artifact_source`` for to get the alpha mask
            of the artifacts to blend them with ``intensities``.

        deformation_strength (``int``, optional):

            Strength of the slice deformation in voxels, used if
            ``prob_deform`` > 0. The deformation models a fold by shifting the
            section contents towards a randomly oriented line in the section.
            The line itself will be drawn with a value of 0.

        axis (``int``, optional):

            Along which axis sections are cut.
    """

    def __init__(
        self,
        intensities,
        prob_missing=0.05,
        prob_low_contrast=0.05,
        prob_artifact=0.0,
        prob_deform=0.0,
        contrast_scale=0.1,
        artifact_source=None,
        artifacts=None,
        artifacts_mask=None,
        deformation_strength=20,
        axis=0,
    ):
        self.intensities = intensities
        self.prob_missing = prob_missing
        self.prob_low_contrast = prob_low_contrast
        self.prob_artifact = prob_artifact
        self.prob_deform = prob_deform
        self.contrast_scale = contrast_scale
        self.artifact_source = artifact_source
        self.artifacts = artifacts
        self.artifacts_mask = artifacts_mask
        self.deformation_strength = deformation_strength
        self.axis = axis

    def setup(self):
        if self.artifact_source is not None:
            self.artifact_source.setup()

    def teardown(self):
        if self.artifact_source is not None:
            self.artifact_source.teardown()

    # send roi request to data-source upstream
    def prepare(self, request):
        deps = BatchRequest()

        # we prepare the augmentations, by determining which slices
        # will be augmented by which method
        # If one of the slices is augmented with 'deform',
        # we prepare these trafos already
        # and request a bigger roi from upstream

        prob_missing_threshold = self.prob_missing
        prob_low_contrast_threshold = prob_missing_threshold + self.prob_low_contrast
        prob_artifact_threshold = prob_low_contrast_threshold + self.prob_artifact
        prob_deform_slice = prob_artifact_threshold + self.prob_deform

        spec = request[self.intensities].copy()
        roi = spec.roi
        logger.debug("downstream request ROI is %s" % roi)
        raw_voxel_size = self.spec[self.intensities].voxel_size

        # store the mapping slice to augmentation type in a dict
        self.slice_to_augmentation = {}
        # store the transformations for deform slice
        self.deform_slice_transformations = {}
        for c in range((roi / raw_voxel_size).shape[self.axis]):
            r = random.random()

            if r < prob_missing_threshold:
                logger.debug("Zero-out " + str(c))
                self.slice_to_augmentation[c] = "zero_out"

            elif r < prob_low_contrast_threshold:
                logger.debug("Lower contrast " + str(c))
                self.slice_to_augmentation[c] = "lower_contrast"

            elif r < prob_artifact_threshold:
                logger.debug("Add artifact " + str(c))
                self.slice_to_augmentation[c] = "artifact"

            elif r < prob_deform_slice:
                logger.debug("Add deformed slice " + str(c))
                self.slice_to_augmentation[c] = "deformed_slice"
                # get the shape of a single slice
                slice_shape = (roi / raw_voxel_size).shape
                slice_shape = slice_shape[: self.axis] + slice_shape[self.axis + 1 :]
                self.deform_slice_transformations[c] = self.__prepare_deform_slice(
                    slice_shape
                )

        # prepare transformation and
        # request bigger upstream roi for deformed slice
        if "deformed_slice" in self.slice_to_augmentation.values():
            # create roi sufficiently large to feed deformation
            logger.debug("before growth: %s" % spec.roi)
            growth = Coordinate(
                tuple(
                    0
                    if d == self.axis
                    else raw_voxel_size[d] * self.deformation_strength
                    for d in range(spec.roi.dims)
                )
            )
            logger.debug("growing request by %s" % str(growth))
            source_roi = roi.grow(growth, growth)

            # update request ROI to get all voxels necessary to perfrom
            # transformation
            spec.roi = source_roi
            logger.debug("upstream request roi is %s" % spec.roi)

        deps[self.intensities] = spec

    def process(self, batch, request):
        assert batch.get_total_roi().dims == 3, "defectaugment works on 3d batches only"

        raw = batch.arrays[self.intensities]
        raw_voxel_size = self.spec[self.intensities].voxel_size

        for c, augmentation_type in self.slice_to_augmentation.items():
            section_selector = tuple(
                slice(None if d != self.axis else c, None if d != self.axis else c + 1)
                for d in range(raw.spec.roi.dims)
            )

            if augmentation_type == "zero_out":
                raw.data[section_selector] = 0

            elif augmentation_type == "low_contrast":
                section = raw.data[section_selector]

                mean = section.mean()
                section -= mean
                section *= self.contrast_scale
                section += mean

                raw.data[section_selector] = section

            elif augmentation_type == "artifact":
                section = raw.data[section_selector]

                alpha_voxel_size = self.artifact_source.spec[
                    self.artifacts_mask
                ].voxel_size

                assert raw_voxel_size == alpha_voxel_size, (
                    "Can only alpha blend RAW with "
                    "ALPHA_MASK if both have the same "
                    "voxel size"
                )

                artifact_request = BatchRequest()
                artifact_request.add(
                    self.artifacts,
                    Coordinate(section.shape) * raw_voxel_size,
                    voxel_size=raw_voxel_size,
                )
                artifact_request.add(
                    self.artifacts_mask,
                    Coordinate(section.shape) * alpha_voxel_size,
                    voxel_size=raw_voxel_size,
                )
                logger.debug("Requesting artifact batch %s", artifact_request)

                artifact_batch = self.artifact_source.request_batch(artifact_request)
                artifact_alpha = artifact_batch.arrays[self.artifacts_mask].data
                artifact_raw = artifact_batch.arrays[self.artifacts].data

                assert artifact_alpha.dtype == np.float32
                assert artifact_alpha.min() >= 0.0
                assert artifact_alpha.max() <= 1.0

                raw.data[section_selector] = (
                    section * (1.0 - artifact_alpha) + artifact_raw * artifact_alpha
                )

            elif augmentation_type == "deformed_slice":
                section = raw.data[section_selector].squeeze()

                # set interpolation to cubic, spec interploatable is true, else to 0
                interpolation = 3 if self.spec[self.intensities].interpolatable else 0

                # load the deformation fields that were prepared for this slice
                flow_x, flow_y, line_mask = self.deform_slice_transformations[c]

                # apply the deformation fields
                shape = section.shape
                section = map_coordinates(
                    section, (flow_y, flow_x), mode="constant", order=interpolation
                ).reshape(shape)

                # things can get smaller than 0 at the boundary, so we clip
                section = np.clip(section, 0.0, 1.0)

                # zero-out data below the line mask
                section[line_mask] = 0.0

                raw.data[section_selector] = section

        # in case we needed to change the ROI due to a deformation augment,
        # restore original ROI and crop the array data
        if "deformed_slice" in self.slice_to_augmentation.values():
            old_roi = request[self.intensities].roi
            logger.debug("resetting roi to %s" % old_roi)
            crop = tuple(
                slice(None)
                if d == self.axis
                else slice(self.deformation_strength, -self.deformation_strength)
                for d in range(raw.spec.roi.dims)
            )
            raw.data = raw.data[crop]
            raw.spec.roi = old_roi

    def __prepare_deform_slice(self, slice_shape):
        # grow slice shape by 2 x deformation strength
        grow_by = 2 * self.deformation_strength
        shape = (slice_shape[0] + grow_by, slice_shape[1] + grow_by)

        # randomly choose fixed x or fixed y with p = 1/2
        fixed_x = random.random() < 0.5
        if fixed_x:
            x0, y0 = 0, np.random.randint(1, shape[1] - 2)
            x1, y1 = shape[0] - 1, np.random.randint(1, shape[1] - 2)
        else:
            x0, y0 = np.random.randint(1, shape[0] - 2), 0
            x1, y1 = np.random.randint(1, shape[0] - 2), shape[1] - 1

        ## generate the mask of the line that should be blacked out
        line_mask = np.zeros(shape, dtype="bool")
        rr, cc = line(x0, y0, x1, y1)
        line_mask[rr, cc] = 1

        # generate vectorfield pointing towards the line to compress the image
        # first we get the unit vector representing the line
        line_vector = np.array([x1 - x0, y1 - y0], dtype="float32")
        line_vector /= np.linalg.norm(line_vector)
        # next, we generate the normal to the line
        normal_vector = np.zeros_like(line_vector)
        normal_vector[0] = -line_vector[1]
        normal_vector[1] = line_vector[0]

        # make meshgrid
        x, y = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]))
        # generate the vector field
        flow_x, flow_y = np.zeros(shape), np.zeros(shape)

        # find the 2 components where coordinates are bigger / smaller than the line
        # to apply normal vector in the correct direction
        components, n_components = label(np.logical_not(line_mask).view("uint8"))
        assert n_components == 2, "%i" % n_components
        neg_val = components[0, 0] if fixed_x else components[-1, -1]
        pos_val = components[-1, -1] if fixed_x else components[0, 0]

        flow_x[components == pos_val] = self.deformation_strength * normal_vector[1]
        flow_y[components == pos_val] = self.deformation_strength * normal_vector[0]
        flow_x[components == neg_val] = -self.deformation_strength * normal_vector[1]
        flow_y[components == neg_val] = -self.deformation_strength * normal_vector[0]

        # generate the flow fields
        flow_x, flow_y = (x + flow_x).reshape(-1, 1), (y + flow_y).reshape(-1, 1)

        # dilate the line mask
        line_mask = binary_dilation(line_mask, iterations=10)

        return flow_x, flow_y, line_mask



from typing import List
import logging

import numpy as np

from gunpowder.array import ArrayKey
from gunpowder.batch import Batch
from gunpowder.batch_request import BatchRequest
from .batch_filter import BatchFilter


logger = logging.getLogger(__name__)


class Unsqueeze(BatchFilter):
    """Unsqueeze a batch at a given axis

    Args:
        arrays (List[ArrayKey]): ArrayKeys to unsqueeze.
        axis: Position where the new axis is placed, defaults to 0.
    """

    def __init__(self, arrays: List[ArrayKey], axis: int = 0):
        self.arrays = arrays
        self.axis = axis

    def setup(self):
        self.enable_autoskip()
        for array in self.arrays:
            self.updates(array, self.spec[array].copy())

    def prepare(self, request):
        deps = BatchRequest()
        for array in self.arrays:
            if array in request:
                deps[array] = request[array].copy()
        return deps

    def process(self, batch, request):
        outputs = Batch()
        for array in self.arrays:
            if array in batch:
                if not batch[array].spec.nonspatial:
                    spatial_dims = request[array].roi.dims
                    if self.axis > batch[array].data.ndim - spatial_dims:
                        raise ValueError(
                            (
                                f"Unsqueeze.axis={self.axis} not permitted. "
                                "Unsqueeze only supported for "
                                "non-spatial dimensions of Array."
                            )
                        )

                outputs[array] = batch[array]
                outputs[array].data = np.expand_dims(batch[array].data, self.axis)
        return outputs



import sys

PY2 = sys.version_info[0] == 2

if PY2:
    binary_type = str
else:
    binary_type = bytes


def ensure_str(s):
    if PY2:
        if isinstance(s, buffer):
            s = str(s)
    else:
        if isinstance(s, memoryview):
            s = s.tobytes()
        if isinstance(s, binary_type):
            s = s.decode("ascii")
    return s



from __future__ import absolute_import

from .add_affinities import AddAffinities
from .astype import AsType
from .balance_labels import BalanceLabels
from .batch_filter import BatchFilter
from .batch_provider import BatchProvider
from .crop import Crop
from .csv_points_source import CsvPointsSource
from .daisy_request_blocks import DaisyRequestBlocks
from .defect_augment import DefectAugment
from .downsample import DownSample
from .dvid_source import DvidSource
from .deform_augment import DeformAugment
from .elastic_augment import ElasticAugment
from .exclude_labels import ExcludeLabels
from .graph_source import GraphSource
from .grow_boundary import GrowBoundary
from .hdf5_source import Hdf5Source
from .hdf5_write import Hdf5Write
from .intensity_augment import IntensityAugment
from .intensity_scale_shift import IntensityScaleShift
from .iterate_locations import IterateLocations
from .klb_source import KlbSource
from .merge_provider import MergeProvider
from .noise_augment import NoiseAugment
from .normalize import Normalize
from .pad import Pad
from .precache import PreCache
from .print_profiling_stats import PrintProfilingStats
from .random_location import RandomLocation
from .random_provider import RandomProvider
from .rasterize_graph import RasterizationSettings, RasterizeGraph
from .reject import Reject
from .renumber_connected_components import RenumberConnectedComponents
from .resample import Resample
from .scan import Scan
from .shift_augment import ShiftAugment
from .simple_augment import SimpleAugment
from .snapshot import Snapshot
from .specified_location import SpecifiedLocation
from .squeeze import Squeeze
from .stack import Stack
from .unsqueeze import Unsqueeze
from .upsample import UpSample
from .zarr_source import ZarrSource
from .zarr_write import ZarrWrite



import logging
import numpy as np
from gunpowder.ext import jax
from gunpowder.ext import jnp
import pickle
import os

from gunpowder.array import ArrayKey, Array
from gunpowder.array_spec import ArraySpec
from gunpowder.ext import tensorboardX, NoSuchModule
from gunpowder.nodes.generic_train import GenericTrain
from gunpowder.jax import GenericJaxModel

from typing import Dict, Union, Optional, Any


logger = logging.getLogger(__name__)


class Train(GenericTrain):
    """JAX implementation of :class:`gunpowder.nodes.GenericTrain`.

    Args:

        model (subclass of ``gunpowder.jax.GenericJaxModel``):

            The model to train. This model encapsulates the forward model,
            loss, and optimizer.

        inputs (``dict``, ``string`` -> Union[np.ndarray, ArrayKey]):

            Dictionary from the names of input tensors expected by the
            ``train_step`` method to array keys or ndarray.

        outputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of tensors in the network to array
            keys. If the key is a string, the tensor will be retrieved
            by checking the model for an attribute with they key as its name.
            If the key is an integer, it is interpreted as a tuple index of
            the outputs of the network.
            New arrays will be generated by this node for each entry (if
            requested downstream).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (at the moment only
            ``output``). This is useful to set the ``voxel_size``, for example,
            if they differ from the voxel size of the input arrays. Only fields
            that are not ``None`` in the given :class:`ArraySpec` will be used.

        checkpoint_basename (``string``, optional):

            The basename used for checkpoint files. Defaults to ``model``.

        save_every (``int``, optional):

            After how many iterations to create a checkpoint to store the
            learnt weights.

        keep_n_checkpoints (``int``, optional):

            Number of checkpoints to keep. Node will attempt to delete older
            checkpoints. Default is `None` (no deletion).

        log_dir (``string``, optional):

            Directory for saving tensorboard summaries.

        log_every (``int``, optional):

            After how many iterations to write out tensorboard summaries.

        spawn_subprocess (``bool``, optional):

            Whether to run the ``train_step`` in a separate process. Default is
            false.

        n_devices (``int``, optional):

            Number of GPU devices to train on concurrently using `jax.pmap`. If
            `None`, the number of available GPUs will be automatically detected
            and used.

        validate_fn (function -> Union[``float``, (``dict``, ``string`` -> ``float``)] , optional):

            Function to run validation on, which should has the form of

                def validate_fn(model, params)

            where `model` is the same provided `GenericJaxModel` model and
            `params` is the parameter of this model, and returns either a
            ``float`` (one loss) or a dictionary of losses to record in
            tensorboard.

        validate_every (``int``, optional):

            After how many iterations to run `validate_fn`.
    """

    def __init__(
        self,
        model: GenericJaxModel,
        inputs: Dict[str, Union[np.ndarray, ArrayKey]],
        outputs: Dict[Union[int, str], ArrayKey],
        gradients: Dict[Union[int, str], ArrayKey] = {},
        array_specs: Optional[Dict[ArrayKey, ArraySpec]] = None,
        checkpoint_basename: str = "model",
        save_every: int = 2000,
        keep_n_checkpoints: Optional[int] = None,
        log_dir: Optional[str] = None,
        log_every: int = 1,
        spawn_subprocess: bool = False,
        n_devices: Optional[int] = None,
        validate_fn=None,
        validate_every=None,
    ):
        # not yet implemented
        gradients = gradients

        super(Train, self).__init__(
            inputs, outputs, gradients, array_specs, spawn_subprocess=spawn_subprocess
        )

        self.model = model
        self.checkpoint_basename = checkpoint_basename
        self.save_every = save_every
        if n_devices is None:
            n_devices = jax.local_device_count()  # autodetect available GPUs
        self.n_devices = n_devices
        self.local_devices = jax.local_devices()
        self.keep_n_checkpoints = keep_n_checkpoints

        self.iteration = 0

        if not isinstance(tensorboardX, NoSuchModule) and log_dir is not None:
            self.summary_writer = tensorboardX.SummaryWriter(log_dir)
            self.log_every = log_every
        else:
            self.summary_writer = None
            if log_dir is not None:
                logger.warning("log_dir given, but tensorboardX is not installed")

        self.intermediate_layers: dict[ArrayKey, Any] = {}

        self.validate_fn = validate_fn
        self.validate_every = validate_every

    def replicate_params(self, params):
        return jax.tree_map(lambda x: jnp.array([x] * self.n_devices), params)

    def start(self):
        checkpoint, self.iteration = self._get_latest_checkpoint(
            self.checkpoint_basename
        )

        if checkpoint is not None:
            logger.info("Resuming training from iteration %d", self.iteration)

            with open(checkpoint, "rb") as f:
                self.model_params = pickle.load(f)
                if self.n_devices > 1:
                    self.model_params = self.replicate_params(self.model_params)
        else:
            logger.info("Starting training from scratch")
            self.model_params = None

    def split_inputs(self, inputs):
        for k, arr in inputs.items():
            assert arr.shape[0] % self.n_devices == 0, (
                f"Batch size should be evenly divisible by the number of "
                f"devices. Input array shape is {arr.shape} but n_device is"
                f" {self.n_devices}"
            )
            inputs[k] = arr.reshape(
                self.n_devices, arr.shape[0] // self.n_devices, *arr.shape[1:]
            )
            inputs[k] = [x for x in inputs[k]]  # make a sequence for put_sharded
        return inputs

    def unstack_device_outputs(self, outputs):
        for k, arr in outputs.items():
            outputs[k] = arr.reshape(arr.shape[0] * arr.shape[1], *arr.shape[2:])
        return outputs

    def train_step(self, batch, request):
        inputs = self.__collect_provided_inputs(batch)
        if self.n_devices > 1:
            inputs = self.split_inputs(inputs)

        # put to device for max performance
        if self.n_devices > 1:
            for k, v in inputs.items():
                inputs[k] = jax.device_put_sharded(v, jax.local_devices())
        else:
            for k, v in inputs.items():
                inputs[k] = jax.device_put(v)

        # initialize model if necessary
        if self.model_params is None:
            # Use the random seed of first request to initialize model's weight
            rng = jax.random.PRNGKey(request.random_seed)
            if self.n_devices > 1:
                rng = jnp.broadcast_to(rng, (self.n_devices,) + rng.shape)
                self.model_params = jax.pmap(self.model.initialize)(rng, inputs)
            else:
                self.model_params = self.model.initialize(rng, inputs)

        requested_outputs = self.__collect_requested_outputs(request)

        if self.n_devices > 1:
            self.model_params, outputs, loss = jax.pmap(
                self.model.train_step,
                axis_name="num_devices",
                donate_argnums=(0,),
                static_broadcasted_argnums=(2,),
            )(self.model_params, inputs, True)
            loss = loss.mean()
            outputs = self.unstack_device_outputs(outputs)  # stack by batch
        else:
            self.model_params, outputs, loss = jax.jit(
                self.model.train_step, donate_argnums=(0,), static_argnums=(2,)
            )(self.model_params, inputs, False)

        logger.debug("model outputs: %s", {k: v.shape for k, v in outputs.items()})

        # add requested model outputs to batch
        for array_key, array_name in requested_outputs.items():
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(outputs[array_name], spec)

        if isinstance(loss, dict):
            total = 0.0
            for k, v in loss.items():
                total += v
            batch.loss = total
        else:
            batch.loss = loss
        self.iteration += 1
        batch.iteration = self.iteration

        if batch.iteration % self.save_every == 0:
            checkpoint_name = self._checkpoint_name(
                self.checkpoint_basename, batch.iteration
            )

            logger.info("Creating checkpoint %s", checkpoint_name)

            model_state = self.model_params
            if self.n_devices > 1:
                # get only a single copy of param for saving
                model_state = jax.tree_map(lambda x: x[0], model_state)
            with open(checkpoint_name, "wb") as f:
                pickle.dump(model_state, f)

            if self.keep_n_checkpoints:
                checkpoint_name = self._checkpoint_name(
                    self.checkpoint_basename,
                    batch.iteration - self.keep_n_checkpoints * self.save_every,
                )
                try:
                    os.remove(checkpoint_name)
                    logger.info("Removed checkpoint %s", checkpoint_name)
                except FileNotFoundError:
                    pass

        if self.summary_writer and batch.iteration % self.log_every == 0:
            if isinstance(loss, dict):
                for k, v in loss.items():
                    self.summary_writer.add_scalar(k, v, batch.iteration)
            else:
                self.summary_writer.add_scalar("loss", loss, batch.iteration)

        # run validate
        if self.validate_fn is not None and batch.iteration % self.validate_every == 0:
            val_ret = self.validate_fn(self.model, self.model_params)
            if isinstance(val_ret, dict):
                for k, v in val_ret.items():
                    self.summary_writer.add_scalar(k, v, batch.iteration)
            else:
                self.summary_writer.add_scalar("validate", val_ret, batch.iteration)

    def __collect_requested_outputs(self, request):
        array_outputs = {}

        for output_name, array_key in self.outputs.items():
            if array_key in request:
                array_outputs[array_key] = output_name

        return array_outputs

    def __collect_provided_inputs(self, batch):
        return self.__collect_provided_arrays(self.inputs, batch)

    def __collect_provided_arrays(self, reference, batch):
        arrays = {}

        for array_name, array_key in reference.items():
            if isinstance(array_key, ArrayKey):
                msg = f"batch does not contain {array_key}, array {array_name} will not be set"
                if array_key in batch.arrays:
                    arrays[array_name] = batch.arrays[array_key].data
                else:
                    logger.warn(msg)
            elif isinstance(array_key, np.ndarray):
                arrays[array_name] = array_key
            else:
                raise Exception(
                    "Unknown network array key {}, can't be given to "
                    "network".format(array_key)
                )

        return arrays



import logging

from .batch_filter import BatchFilter
from gunpowder.profiling import Timing, TimingSummary, ProfilingStats

logger = logging.getLogger(__name__)


class PrintProfilingStats(BatchFilter):
    """Print profiling information about nodes upstream of this node in the DAG.

    The output also includes a ``TOTAL`` section, which shows the wall-time
    spent in the upstream and downstream passes. For the downstream pass, this
    information is not available in the first iteration, since the request-batch
    cycle is not completed, yet.

    Args:

        every (``int``):

            Collect statistics about that many batch requests and show min,
            max, mean, and median runtimes.
    """

    def __init__(self, every=1):
        self.every = every
        self.n = 0
        self.accumulated_stats = ProfilingStats()
        self.__upstream_timing = Timing(self)
        self.__upstream_timing_summary = TimingSummary()
        self.__downstream_timing = Timing(self)
        self.__downstream_timing_summary = TimingSummary()

    def prepare(self, request):
        self.__downstream_timing.stop()
        # skip the first one, where we don't know how much time we spent
        # downstream
        if self.__downstream_timing.elapsed() > 0:
            self.__downstream_timing_summary.add(self.__downstream_timing)
            self.__downstream_timing = Timing(self)

        self.__upstream_timing.start()

        deps = request
        return deps

    def process(self, batch, request):
        self.__upstream_timing.stop()
        self.__upstream_timing_summary.add(self.__upstream_timing)
        self.__upstream_timing = Timing(self)

        self.__downstream_timing.start()

        self.n += 1
        print_stats = self.n % self.every == 0

        self.accumulated_stats.merge_with(batch.profiling_stats)

        if not print_stats:
            return

        stats = "\n"
        stats += "Profiling Stats\n"
        stats += "===============\n"
        stats += "\n"
        stats += "NODE".ljust(20)
        stats += "METHOD".ljust(10)
        stats += "COUNTS".ljust(10)
        stats += "MIN".ljust(10)
        stats += "MAX".ljust(10)
        stats += "MEAN".ljust(10)
        stats += "MEDIAN".ljust(10)
        stats += "\n"

        summaries = list(self.accumulated_stats.get_timing_summaries().items())
        summaries.sort()

        for (node_name, method_name), summary in summaries:
            if summary.counts() > 0:
                stats += node_name[:19].ljust(20)
                stats += (
                    method_name[:19].ljust(10) if method_name is not None else " " * 10
                )
                stats += ("%d" % summary.counts())[:9].ljust(10)
                stats += ("%.2f" % summary.min())[:9].ljust(10)
                stats += ("%.2f" % summary.max())[:9].ljust(10)
                stats += ("%.2f" % summary.mean())[:9].ljust(10)
                stats += ("%.2f" % summary.median())[:9].ljust(10)
                stats += "\n"

        stats += "\n"
        stats += "TOTAL"
        stats += "\n"

        for phase, summary in zip(
            ["upstream", "downstream"],
            [self.__upstream_timing_summary, self.__downstream_timing_summary],
        ):
            if summary.counts() > 0:
                stats += phase[:19].ljust(30)
                stats += ("%d" % summary.counts())[:9].ljust(10)
                stats += ("%.2f" % summary.min())[:9].ljust(10)
                stats += ("%.2f" % summary.max())[:9].ljust(10)
                stats += ("%.2f" % summary.mean())[:9].ljust(10)
                stats += ("%.2f" % summary.median())[:9].ljust(10)
                stats += "\n"

        stats += "\n"

        logger.info(stats)

        # reset summaries
        self.accumulated_stats = ProfilingStats()
        self.__upstream_timing_summary = TimingSummary()
        self.__downstream_timing_summary = TimingSummary()



import logging
import multiprocessing
import time

from gunpowder.nodes.batch_filter import BatchFilter
from gunpowder.producer_pool import ProducerPool, WorkersDied, NoResult
from gunpowder.array import ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.batch_request import BatchRequest
from queue import Full

logger = logging.getLogger(__name__)


class PredictProcessDied(Exception):
    pass


class GenericPredict(BatchFilter):
    """Generic predict node to add predictions of a trained network to each each
    batch that passes through. This node alone does nothing and should be
    subclassed for concrete implementations.

    Args:

        inputs (dict): Dictionary from names of input layers in the network to
            :class:``ArrayKey`` or batch attribute name as string.

        outputs (dict): Dictionary from the names of output layers in the
            network to :class:``ArrayKey``. New arrays will be generated by
            this node for each entry (if requested downstream).

        array_specs (dict, optional): An optional dictionary of
            :class:`ArrayKey` to :class:`ArraySpec` to set the array specs
            generated arrays (``outputs`` and ``gradients``). This is useful
            to set the ``voxel_size``, for example, if they differ from the
            voxel size of the input arrays. Only fields that are not ``None``
            in the given :class:`ArraySpec` will be used.

        spawn_subprocess (bool, optional): Whether to run ``predict`` in a
            separate process. Default is false.
    """

    def __init__(self, inputs, outputs, array_specs=None, spawn_subprocess=False):
        self.initialized = False
        self.inputs = inputs
        self.outputs = outputs
        self.array_specs = {} if array_specs is None else array_specs
        self.spawn_subprocess = spawn_subprocess
        self.timer_start = None

    def setup(self):
        # get common voxel size of inputs, or None if they differ
        common_voxel_size = None
        for key in self.inputs.values():
            if not isinstance(key, ArrayKey):
                continue

            voxel_size = self.spec[key].voxel_size

            if common_voxel_size is None:
                common_voxel_size = voxel_size
            elif common_voxel_size != voxel_size:
                common_voxel_size = None
                break

        # announce provided outputs
        for key in self.outputs.values():
            if key in self.array_specs:
                spec = self.array_specs[key].copy()
            else:
                spec = ArraySpec()

            if spec.voxel_size is None and not spec.nonspatial:
                assert common_voxel_size is not None, (
                    "There is no common voxel size of the inputs, and no "
                    "ArraySpec has been given for %s that defines "
                    "voxel_size." % key
                )

                spec.voxel_size = common_voxel_size

            if spec.interpolatable is None:
                # default for predictions
                spec.interpolatable = False

            self.provides(key, spec)

        if self.spawn_subprocess:
            # start prediction as a producer pool, so that we can gracefully
            # exit if anything goes wrong
            self.worker = ProducerPool([self._produce_predict_batch], queue_size=1)
            self.batch_in = multiprocessing.Queue(maxsize=1)
            self.batch_in_lock = multiprocessing.Lock()
            self.batch_out_lock = multiprocessing.Lock()
            self.worker.start()

    def teardown(self):
        if self.spawn_subprocess:
            # signal "stop"
            try:
                self.batch_in.put((None, None), timeout=2)
            except Full:
                # worker process might be stopped already
                pass
            try:
                self.worker.get(timeout=2)
            except NoResult:
                pass
            self.worker.stop()
        else:
            self.stop()

    def prepare(self, request):
        if not self.initialized and not self.spawn_subprocess:
            self.start()
            self.initialized = True

        deps = BatchRequest()
        for key in self.inputs.values():
            deps[key] = request[key]
        return deps

    def process(self, batch, request):
        if self.spawn_subprocess:
            start = time.time()
            self.batch_in_lock.acquire()
            logger.debug("waited for batch in lock for %.3fs", time.time() - start)
            start = time.time()
            self.batch_in.put((batch, request))
            logger.debug("queued batch for %.3fs", time.time() - start)

            start = time.time()
            with self.batch_out_lock:
                logger.debug("waited for batch out lock for %.3fs", time.time() - start)

                start = time.time()
                self.batch_in_lock.release()
                logger.debug("released batch in lock for %.3fs", time.time() - start)
                try:
                    start = time.time()
                    out = self.worker.get()
                    logger.debug("retreived batch for %.3fs", time.time() - start)
                except WorkersDied:
                    raise PredictProcessDied()

            for array_key in self.outputs.values():
                if array_key in request:
                    batch.arrays[array_key] = out.arrays[array_key]

        else:
            self.predict(batch, request)

    def start(self):
        """To be implemented in subclasses.

        This method will be called before the first call to :fun:`predict`,
        from the same process that :fun:`predict` will be called from. Use
        this to initialize your model and hardware.
        """
        pass

    def predict(self, batch, request):
        """To be implemented in subclasses.

        In this method, an implementation should predict arrays on the given
        batch. Output arrays should be created according to the given request
        and added to ``batch``."""
        raise NotImplementedError("Class %s does not implement 'predict'" % self.name())

    def stop(self):
        """To be implemented in subclasses.

        This method will be called after the last call to :fun:`predict`,
        from the same process that :fun:`predict` will be called from. Use
        this to tear down your model and free training hardware.
        """
        pass

    def _produce_predict_batch(self):
        """Process one batch."""

        if not self.initialized:
            self.start()
            self.initialized = True

        self.time_out = 0
        if self.timer_start is not None:
            self.time_out = time.time() - self.timer_start

        self.timer_start = time.time()
        batch, request = self.batch_in.get()
        self.time_in = time.time() - self.timer_start

        # stop signal
        if batch is None:
            self.stop()
            return None

        self.timer_start = time.time()
        self.predict(batch, request)
        self.time_predict = time.time() - self.timer_start
        self.timer_start = time.time()

        logger.debug(
            "batch in: %.3fs, predict: %.3fs, batch out: %.3fs",
            self.time_in,
            self.time_predict,
            self.time_out,
        )

        return batch



from gunpowder.nodes.batch_filter import BatchFilter


class ZeroOutConstSections(BatchFilter):
    """Every z-section that has constant values only will be set to 0.

    This is to handle blank (missing) sections in a less invasive way: Instead
    of leaving them at -1 (which is "black", the lowest possible input to the
    CNN), 0 ("gray") might be easier to ignore.

    For that you should call this filter after you are done with all other
    intensity manipulations.
    """

    def __init__(self, intensities):
        self.intensities = intensities

    def process(self, batch, request):
        assert batch.get_total_roi().dims == 3, "This filter only works on 3D data."

        raw = batch.arrays[self.intensities]

        for z in range(
            (raw.spec.roi / self.spec[self.intensities].voxel_size).get_shape()[0]
        ):
            if raw.data[z].min() == raw.data[z].max():
                raw.data[z] = 0



import logging
import multiprocessing
import ctypes
from gunpowder.ext import tensorflow as tf
from gunpowder.freezable import Freezable

logger = logging.getLogger(__name__)


class LocalServer(Freezable):
    """Wrapper around ``tf.train.Server`` to create a local server on-demand.

    This class is necessary because tensorflow's GPU support should not be
    initialized before forking processes (the CUDA driver needs to be
    initialized in each process separately, not in the main process and then
    forked). Creating a ``tf.train.Server`` initializes GPU support, however.
    With this wrapper, server creating can be delayed until a GPU process
    creates a ``tf.Session``::

        session = tf.Session(target=LocalServer.get_target())
    """

    __target = multiprocessing.Array(ctypes.c_char, b" " * 256)
    __server = None

    @staticmethod
    def get_target():
        """Get the target string of this tensorflow server to connect a
        ``tf.Session()``. This will start the server, if it is not running
        already.
        """

        with LocalServer.__target.get_lock():
            target = LocalServer.__target.value

            if target == b" " * 256:
                logger.info("Creating local tensorflow server")
                LocalServer.__server = tf.train.Server.create_local_server()
                target = LocalServer.__server.target
                if not isinstance(target, bytes):
                    target = target.encode("ascii")
                logger.info("Server running at %s", target)
            else:
                logger.info("Server already running at %s", target)

            LocalServer.__target.value = target

        return target



import numpy as np
import logging
from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.nodes.batch_provider import BatchProvider
from gunpowder.graph import Node, Graph
from gunpowder.graph_spec import GraphSpec
from gunpowder.profiling import Timing
from gunpowder.roi import Roi

logger = logging.getLogger(__name__)


class CsvPointsSource(BatchProvider):
    """Read a set of points from a comma-separated-values text file. Each line
    in the file represents one point, e.g. z y x (id)

    Args:

        filename (``string``):

            The file to read from.

        points (:class:`GraphKey`):

            The key of the points set to create.

        points_spec (:class:`GraphSpec`, optional):

            An optional :class:`GraphSpec` to overwrite the points specs
            automatically determined from the CSV file. This is useful to set
            the :class:`Roi` manually.

        scale (scalar or array-like):

            An optional scaling to apply to the coordinates of the points read
            from the CSV file. This is useful if the points refer to voxel
            positions to convert them to world units.

        ndims (``int``):

            If ``ndims`` is None, all values in one line are considered as the
            location of the point. If positive, only the first ``ndims`` are used.
            If negative, all but the last ``-ndims`` are used.

         id_dim (``int``):

            Each line may optionally contain an id for each point. This parameter
            specifies its location, has to come after the position values.
    """

    def __init__(
        self, filename, points, points_spec=None, scale=None, ndims=None, id_dim=None
    ):
        self.filename = filename
        self.points = points
        self.points_spec = points_spec
        self.scale = scale
        self.ndims = ndims
        self.id_dim = id_dim
        self.data = None

    def setup(self):
        self._parse_csv()

        if self.points_spec is not None:
            self.provides(self.points, self.points_spec)
            return

        min_bb = Coordinate(np.floor(np.amin(self.data[:, : self.ndims], 0)))
        max_bb = Coordinate(np.ceil(np.amax(self.data[:, : self.ndims], 0)) + 1)

        roi = Roi(min_bb, max_bb - min_bb)

        self.provides(self.points, GraphSpec(roi=roi))

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        min_bb = request[self.points].roi.begin
        max_bb = request[self.points].roi.end

        logger.debug("CSV points source got request for %s", request[self.points].roi)

        point_filter = np.ones((self.data.shape[0],), dtype=bool)
        for d in range(self.ndims):
            point_filter = np.logical_and(point_filter, self.data[:, d] >= min_bb[d])
            point_filter = np.logical_and(point_filter, self.data[:, d] < max_bb[d])

        points_data = self._get_points(point_filter)
        points_spec = GraphSpec(roi=request[self.points].roi.copy())

        batch = Batch()
        batch.graphs[self.points] = Graph(points_data, [], points_spec)

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def _get_points(self, point_filter):
        filtered = self.data[point_filter][:, : self.ndims]

        if self.id_dim is not None:
            ids = self.data[point_filter][:, self.id_dim]
        else:
            ids = np.arange(len(self.data))[point_filter]

        return [Node(id=i, location=p) for i, p in zip(ids, filtered)]

    def _parse_csv(self):
        """Read one point per line. If ``ndims`` is None, all values in one line
        are considered as the location of the point. If positive, only the
        first ``ndims`` are used. If negative, all but the last ``-ndims`` are
        used.
        """

        with open(self.filename, "r") as f:
            self.data = np.array(
                [[float(t.strip(",")) for t in line.split()] for line in f],
                dtype=np.float32,
            )

        if self.ndims is None:
            self.ndims = self.data.shape[1]

        if self.scale is not None:
            self.data[:, : self.ndims] *= self.scale



import numpy as np

from gunpowder.batch_request import BatchRequest

from .batch_filter import BatchFilter


class IntensityAugment(BatchFilter):
    """Randomly scale and shift the values of an intensity array.

    Args:

        array (:class:`ArrayKey`):

            The intensity array to modify.

        scale_min (``float``):
        scale_max (``float``):
        shift_min (``float``):
        shift_max (``float``):

            The min and max of the uniformly randomly drawn scaling and
            shifting values for the intensity augmentation. Intensities are
            changed as::

                a = a.mean() + (a-a.mean())*scale + shift

        z_section_wise (``bool``):

            Perform the augmentation z-section wise. Requires 3D arrays and
            assumes that z is the first dimension.

        clip (``bool``):

            Set to False if modified values should not be clipped to [0, 1]
            Disables range check!
    """

    def __init__(
        self,
        array,
        scale_min,
        scale_max,
        shift_min,
        shift_max,
        z_section_wise=False,
        clip=True,
    ):
        self.array = array
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.shift_min = shift_min
        self.shift_max = shift_max
        self.z_section_wise = z_section_wise
        self.clip = clip

    def setup(self):
        self.enable_autoskip()
        self.updates(self.array, self.spec[self.array])

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.array] = request[self.array].copy()
        return deps

    def process(self, batch, request):
        raw = batch.arrays[self.array]

        assert (
            not self.z_section_wise or raw.spec.roi.dims == 3
        ), "If you specify 'z_section_wise', I expect 3D data."
        assert raw.data.dtype == np.float32 or raw.data.dtype == np.float64, (
            "Intensity augmentation requires float types for the raw array (not "
            + str(raw.data.dtype)
            + "). Consider using Normalize before."
        )
        if self.clip:
            assert (
                raw.data.min() >= 0 and raw.data.max() <= 1
            ), "Intensity augmentation expects raw values in [0,1]. Consider using Normalize before."

        if self.z_section_wise:
            for z in range((raw.spec.roi / self.spec[self.array].voxel_size).shape[0]):
                raw.data[z] = self.__augment(
                    raw.data[z],
                    np.random.uniform(low=self.scale_min, high=self.scale_max),
                    np.random.uniform(low=self.shift_min, high=self.shift_max),
                )
        else:
            raw.data = self.__augment(
                raw.data,
                np.random.uniform(low=self.scale_min, high=self.scale_max),
                np.random.uniform(low=self.shift_min, high=self.shift_max),
            )

        # clip values, we might have pushed them out of [0,1]
        if self.clip:
            raw.data[raw.data > 1] = 1
            raw.data[raw.data < 0] = 0

    def __augment(self, a, scale, shift):
        return a.mean() + (a - a.mean()) * scale + shift
